@@ -1,12 +1,92 @@
 /*
- * Copyright (c) 2001, Raphael Manfredi
+ * $Id: matching.c,v 1.31 2004/01/15 23:51:48 wyldfire Exp $
  *
- * Search bins are Copyright (c) 2001, Kenn Brooks Hamm & Raphael Manfredi
+ * Copyright (c) 2001-2003, Raphael Manfredi
+ *
+ * Search bins are Copyright (c) 2001-2003, Kenn Brooks Hamm & Raphael Manfredi
+ *
+ *----------------------------------------------------------------------
+ * This file is part of gtk-gnutella.
+ *
+ *  gtk-gnutella is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+ *
+ *  gtk-gnutella is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with gtk-gnutella; if not, write to the Free Software
+ *  Foundation, Inc.:
+ *      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *----------------------------------------------------------------------
  */
 
+#include "common.h"
+
 #include <ctype.h>
-#include "gnutella.h"
-#include "matching.h"
+
+#include "qrp.h"			/* For qhvec_add() */
+#include "override.h"			/* Must be the last header included */
+
+RCSID("$Id: matching.c,v 1.31 2004/01/15 23:51:48 wyldfire Exp $");
+
+/* FIXME: remove this dependency */
+extern guint32 dbg;
+
+#define WOVEC_DFLT	10				/* Default size of word-vectors */
+
+/*
+ * Masks for mask_hash().
+ */
+
+#define MASK_LETTER(x)		(1 << (x))		/* bits 0 to 25 */
+#define MASK_DIGIT			0x80000000
+
+static zone_t *pat_zone = NULL;		/* Compiled patterns */
+static zone_t *wovec_zone = NULL;	/* Word-vectors of WOVEC_DFLT entries */
+
+
+
+/*
+ * matching_init
+ *
+ * Initialize matching data structures.
+ */
+void matching_init(void)
+{
+	/*
+	 * We don't expect much word vectors to be created.  They are normally
+	 * created and destroyed in the same routine, without any threading
+	 * taking place.
+	 *
+	 * We only allocate word vectors of WOVEC_DFLT entries in the zone.
+	 * If we need to expand that, it will be done through regular malloc().
+	 */
+
+	wovec_zone = zget(WOVEC_DFLT * sizeof(word_vec_t), 2);
+
+	/*
+	 * Patterns are not only used for query matching but also for filters,
+	 * therefore we can expect quite a few to be created at the same time.
+	 */
+
+	pat_zone = zget(sizeof(cpattern_t), 64);
+}
+
+/*
+ * matching_close
+ *
+ * Terminate matching data structures.
+ */
+void matching_close(void)
+{
+	zdestroy(wovec_zone);
+	zdestroy(pat_zone);
+}
 
 /*
  * Search query word splitting.
@@ -20,6 +100,23 @@
  * but also their wanted frequency.
  */
 
+/*
+ * word_vec_zrealloc
+ *
+ * Reallocate a word-vector from the zone into heap memory, to hold `ncount'.
+ */
+static word_vec_t *word_vec_zrealloc(word_vec_t *wv, gint ncount)
+{
+	word_vec_t *nwv = g_malloc(ncount * sizeof(word_vec_t));
+
+	g_assert(ncount > WOVEC_DFLT);
+
+	memcpy(nwv, wv, WOVEC_DFLT * sizeof(word_vec_t));
+	zfree(wovec_zone, wv);
+
+	return nwv;
+}
+
 /*
  * query_make_word_vec
  *
@@ -31,56 +128,106 @@
  * with the pointer to the allocated vector.  If there are no items, there
  * is no vector returned.
  */
-guint query_make_word_vec(guchar *query, word_vec_t **wovec)
+guint query_make_word_vec(const gchar *query_str, word_vec_t **wovec)
 {
 	guint n = 0;
-	GHashTable *seen_word = g_hash_table_new(g_str_hash, g_str_equal);
-	guint nv = 10;
-	word_vec_t *wv = g_malloc(nv * sizeof(word_vec_t));
+	GHashTable *seen_word = NULL;
+	guint nv = WOVEC_DFLT;
+	word_vec_t *wv = zalloc(wovec_zone);
+	const gchar *start = NULL;
+	gchar * const query_dup = g_strdup(query_str);
+	gchar *query;
+	gchar first = TRUE;
 	guchar c;
-	gchar *start = NULL;
 
 	g_assert(wovec != NULL);
 
-	for (;; query++) {
+	for (query = query_dup; /* empty */; query++) {
 		gboolean is_alpha;
-		c = *query;
-		is_alpha = c ? isalnum(c) : FALSE;
+
+		c = *(guchar *) query;
+#ifdef USE_ICU
+		is_alpha = (c != ' ' && c != '\0');
+		/*
+		 * We can't meet other non alpha than space, because the
+		 * string is normalised.
+		 */
+#else
+		is_alpha = isalnum(c);
+#endif
+
 		if (start == NULL) {				/* Not in a word yet */
 			if (is_alpha) start = query;
 		} else {
 			guint np1;
 			if (is_alpha) continue;
 			*query = '\0';
-			/*
-			 * If word already seen in query, it's in the seen_word table.
-			 * The associated value is the index in the vector plus 1.
-			 */
-			np1 = (guint) g_hash_table_lookup(seen_word, (gconstpointer) start);
-			if (np1) wv[np1-1].amount++;
+
+			/* Only create a hash table if there is more than one word. */
+			if (first) 
+				np1 = 0;
 			else {
+				if (seen_word == NULL) {
+					seen_word = g_hash_table_new(g_str_hash, g_str_equal);
+					g_hash_table_insert(seen_word, wv[0].word,
+						GUINT_TO_POINTER(1));
+				}
+
+				/*
+			 	 * If word already seen in query, it's in the seen_word table.
+		 	 	 * The associated value is the index in the vector plus 1.
+		 	 	 */
+
+				np1 = GPOINTER_TO_UINT(
+					g_hash_table_lookup(seen_word, (gconstpointer) start));
+			}
+
+			if (np1--) {
+				wv[np1].amount++;
+				wv[np1].len = query - start;
+			} else {
 				word_vec_t *entry;
 				if (n == nv) {				/* Filled all the slots */
 					nv *= 2;
-					wv = g_realloc(wv, nv * sizeof(word_vec_t));
+					if (n > WOVEC_DFLT)
+						wv = g_realloc(wv, nv * sizeof(word_vec_t));
+					else
+						wv = word_vec_zrealloc(wv, nv);
 				}
 				entry = &wv[n++];
-				entry->word = g_strdup(start);
+				entry->len = query - start;
+				entry->word = walloc(entry->len + 1);	/* For trailing NUL */
+				memcpy(entry->word, start, entry->len + 1); /* Includes NUL */
+				
 				entry->amount = 1;
-				g_hash_table_insert(seen_word, entry->word, (gpointer) n);
+
+				/*
+				 * Delay insertion of first word until we find another one.
+				 * The hash table storing duplicates is not created for
+				 * the first word.  The word entry is saved into `first_word'
+				 * for later insertion, if needed.
+				 */
+
+				if (first)
+					first = FALSE;
+				else { 
+					g_hash_table_insert(seen_word, entry->word,
+						GUINT_TO_POINTER(n));
+				}
 			}
-			*query = c;
 			start = NULL;
 		}
-		if (!c) break;
+
+		if (c == '\0') break;
 	}
 
-	g_hash_table_destroy(seen_word);	/* Key pointers belong to vector */
+	if (NULL != seen_word)
+		g_hash_table_destroy(seen_word);	/* Key pointers belong to vector */
 	if (n)
 		*wovec = wv;
 	else
-		g_free(wv);
-
+		zfree(wovec_zone, wv);
+	g_free(query_dup);
 	return n;
 }
 
@@ -94,9 +241,12 @@ void query_word_vec_free(word_vec_t *wovec, guint n)
 	guint i;
 
 	for (i = 0; i < n; i++)
-		g_free(wovec[i].word);
+		wfree(wovec[i].word, wovec[i].len + 1);
 
-	g_free(wovec);
+	if (n > WOVEC_DFLT)
+		G_FREE_NULL(wovec);
+	else
+		zfree(wovec_zone, wovec);
 }
 
 /*
@@ -116,9 +266,9 @@ void query_word_vec_free(word_vec_t *wovec, guint n)
  *
  * Returns a compiled pattern structure.
  */
-cpattern_t *pattern_compile(guchar *pattern)
+cpattern_t *pattern_compile(gchar *pattern)
 {
-	cpattern_t *p = (cpattern_t *) g_malloc0(sizeof(cpattern_t));
+	cpattern_t *p = (cpattern_t *) zalloc(pat_zone);
 	guint32 plen = strlen(pattern);
 	guint32 *pd = p->delta;
 	gint i;
@@ -134,8 +284,8 @@ cpattern_t *pattern_compile(guchar *pattern)
 
 	plen--;			/* Restore original pattern length */
 
-	for (pd = p->delta, c = pattern, i = 0; i < plen; c++, i++)
-		pd[(guint) *c] = plen - i;
+	for (pd = p->delta, c = (guchar *) pattern, i = 0; i < plen; c++, i++)
+		pd[*c] = plen - i;
 
 	return p;
 }
@@ -147,8 +297,8 @@ cpattern_t *pattern_compile(guchar *pattern)
  */
 void pattern_free(cpattern_t *cpat)
 {
-	g_free(cpat->pattern);
-	g_free(cpat);
+	G_FREE_NULL(cpat->pattern);
+	zfree(pat_zone, cpat);
 }
 
 /*
@@ -162,17 +312,17 @@ void pattern_free(cpattern_t *cpat)
  */
 gchar *pattern_qsearch(
 	cpattern_t *cpat,		/* Compiled pattern */
-	guchar *text,			/* Text we're scanning */
+	gchar *text,			/* Text we're scanning */
 	guint32 tlen,			/* Text length, 0 = compute strlen(text) */
 	guint32 toffset,		/* Offset within text for search start */
 	qsearch_mode_t word)	/* Beginning/whole word matching? */
 {
-	guchar *p;			/* Pointer within string pattern */
-	guchar *t;			/* Pointer within text */
-	guchar *tp;			/* Initial local search text pointer */
+	gchar *p;			/* Pointer within string pattern */
+	gchar *t;			/* Pointer within text */
+	gchar *tp;			/* Initial local search text pointer */
 	guint32 i;			/* Position within pattern string */
-	guchar *start;		/* Start of matching */
-	guchar *end;		/* End of text (first byte after physical end) */
+	gchar *start;		/* Start of matching */
+	gchar *end;			/* End of text (first byte after physical end) */
 	guint32 plen;
 
 	if (!tlen)
@@ -205,7 +355,7 @@ gchar *pattern_qsearch(
 			if (tp == text) {					/* At beginning of text */
 				if (word == qs_begin) return tp;
 				else at_begin = TRUE;
-			} else if (!isalnum(*(tp-1))) {		/* At word boundary */
+			} else if (!isalnum((guchar) *(tp-1))) {	/* At word boundary */
 				if (word == qs_begin) return tp;
 				else at_begin = TRUE;
 			}
@@ -213,17 +363,17 @@ gchar *pattern_qsearch(
 			if (at_begin && word == qs_whole) {
 				if (tp + plen == end)			/* At end of string */
 					return tp;
-				else if (!isalnum(*(tp+plen)))	/* At word boundary after */
-					return tp;
+				else if (!isalnum((guchar) *(tp+plen)))
+					return tp; /* At word boundary after */
 			}
 
 			/* Fall through */
 		}
 
-		tp += cpat->delta[(guint) *(tp + plen)]; /* Continue search there */
+		tp += cpat->delta[(guchar) *(tp + plen)]; /* Continue search there */
 	}
 
-	return (char *) 0;		/* Not found */
+	return NULL;		/* Not found */
 }
 
 /*
@@ -253,51 +403,10 @@ gchar *pattern_qsearch(
 
 #define ST_MIN_BIN_SIZE		4
 
-/*
- * Masks for mask_hash().
- */
-
-#define MASK_LETTER(x)		(1 << (x))		/* bits 0 to 25 */
-#define MASK_DIGIT			0x80000000
-
-/*
- * mask_hash
- *
- * Compute character mask "hash", using one bit per letter of the alphabet,
- * plus one for any digit.
- */
-static guint32 mask_hash(guchar *str) {
-	guchar *s = str;
-	guchar c;
-	guint32 mask = 0;
-
-	while ((c = *s++)) {
-		guchar lc = tolower(c);
-		if (isspace(c))
-			continue;
-		else if (isdigit(c))
-			mask |= MASK_DIGIT;
-		else {
-			gint idx = lc - 'a';
-			if (idx >= 0 && idx < 26)
-				mask |= MASK_LETTER(idx);
-		}
-	}
-
-	return mask;
-}
-
-/* get key of two-char pair */
-__inline__ static gint st_key(search_table_t *table, guchar k[2])
-{
-	return table->index_map[k[0]] * table->nchars +
-		table->index_map[k[1]];
-}
 
 static void destroy_entry(struct st_entry *entry)
 {
-	/* FIXME: make sure this frees everything */
-	g_free(entry->string);
+	G_FREE_NULL(entry->string);
 }
 
 /* initialize a bin */
@@ -326,7 +435,7 @@ static struct st_bin *bin_allocate(void)
  * NOTE: does NOT destroy the st_entry's, since they may be shared */
 static void bin_destroy(struct st_bin *bin)
 {
-	g_free(bin->vals);
+	G_FREE_NULL(bin->vals);
 	bin->vals = 0;
 }
 
@@ -343,18 +452,23 @@ static void bin_insert_item(struct st_bin *bin, struct st_entry *entry)
 /* makes a bin take as little memory as needed */
 static void bin_compact(struct st_bin *bin)
 {
-	g_assert(bin->vals > 0);	/* Or it would not have been allocated */
+	g_assert(bin->vals != NULL);	/* Or it would not have been allocated */
 	bin->vals = g_realloc(bin->vals, bin->nvals * sizeof(bin->vals[0]));
 	bin->nslots = bin->nvals;
 }
 
-/* apply a char map to a string -- returns length of sting */
-static guint map_string(char_map_t map, guchar *string)
+/*
+ * match_map_string
+ *
+ * Apply a char map to a string, inplace.
+ * Returns length of string.
+ */
+guint match_map_string(char_map_t map, gchar *string)
 {
-	guchar *ptr = string;
+	gchar *ptr = string;
 	guchar c;
 
-	while ((c = *ptr))
+	while ((c = (guchar) *ptr))
 		*ptr++ = map[c];
 
 	return ptr - string;
@@ -420,38 +534,72 @@ void st_destroy(search_table_t *table)
 		for (i = 0; i < table->nbins; i++) {
 			if (table->bins[i]) {
 				bin_destroy(table->bins[i]);
-				g_free(table->bins[i]);
+				G_FREE_NULL(table->bins[i]);
 			}
 		}
-		g_free(table->bins);
+		G_FREE_NULL(table->bins);
 		table->bins = 0;
 	}
 
 	if (table->all_entries.vals) {
 		for (i = 0; i < table->all_entries.nvals; i++) {
 			destroy_entry(table->all_entries.vals[i]);
-			g_free(table->all_entries.vals[i]);
+			G_FREE_NULL(table->all_entries.vals[i]);
 		}
 		bin_destroy(&table->all_entries);
 	}
 }
 
+/*
+ * mask_hash
+ *
+ * Compute character mask "hash", using one bit per letter of the alphabet,
+ * plus one for any digit.
+ */
+static guint32 mask_hash(gchar *s) {
+	guchar c;
+	guint32 mask = 0;
+
+	while ((c = (guchar) *s++)) {
+		if (isspace(c))
+			continue;
+		else if (isdigit(c))
+			mask |= MASK_DIGIT;
+		else {
+			gint idx = tolower(c) - 'a';
+			if (idx >= 0 && idx < 26)
+				mask |= MASK_LETTER(idx);
+		}
+	}
+
+	return mask;
+}
+
+/* get key of two-char pair */
+inline gint st_key(search_table_t *table, gchar k[2])
+{
+	return table->index_map[(guchar) k[0]] * table->nchars +
+		table->index_map[(guchar) k[1]];
+}
+
 /* insert an item into the search_table
  * one-char strings are silently ignored */
-void st_insert_item(search_table_t *table, guchar *string, void *data)
+void st_insert_item(search_table_t *table, gchar *string, void *data)
 {
 	gint i;
 	guint len;
 	struct st_entry *entry;
-	GHashTable *seen_keys = g_hash_table_new(g_direct_hash, 0);
+	GHashTable *seen_keys;
 
 	string = g_strdup(string);
 
-	len = map_string(table->fold_map, string);
+	len = match_map_string(table->fold_map, string);
 	if (len < 2) {
-		g_free(string);
+		G_FREE_NULL(string);
 		return;
 	}
+
+	seen_keys = g_hash_table_new(g_direct_hash, 0);
 	
 	entry = g_malloc(sizeof(struct st_entry));
 	entry->string = string;
@@ -462,10 +610,11 @@ void st_insert_item(search_table_t *table, guchar *string, void *data)
 		gint key = st_key(table, string + i);
 
 		/* don't insert item into same bin twice */
-		if (g_hash_table_lookup(seen_keys, (gconstpointer) key))
+		if (g_hash_table_lookup(seen_keys, (gconstpointer)GINT_TO_POINTER(key)))
 			continue;
 
-		g_hash_table_insert(seen_keys, (gpointer) key, (gpointer) 1);
+		g_hash_table_insert(seen_keys, GINT_TO_POINTER(key),
+			GINT_TO_POINTER(1));
 
 		g_assert(key < table->nbins);
 		if (table->bins[key] == NULL)
@@ -479,8 +628,56 @@ void st_insert_item(search_table_t *table, guchar *string, void *data)
 	g_hash_table_destroy(seen_keys);
 }
 
+/* minimize space consumption */
+void st_compact(search_table_t *table)
+{
+	gint i;
+
+	if (!table->all_entries.nvals)
+		return;			/* Nothing in table */
+
+	bin_compact(&table->all_entries);
+	for (i = 0; i < table->nbins; i++)
+		if (table->bins[i])
+			bin_compact(table->bins[i]);
+}
+
 /*
+ * pattern_compile_fast
+ *
+ * Same as pattern_compile(), but the pattern string is NOT duplicated,
+ * and its length is known upon entry.
+ *
+ * NB: there is no pattern_free_fast(), just call zfree() on the result.
+ */
+static cpattern_t *pattern_compile_fast(gchar *pattern, guint32 plen)
+{
+	cpattern_t *p = (cpattern_t *) zalloc(pat_zone);
+	guint32 *pd = p->delta;
+	gint i;
+	guchar *c;
+
+	p->pattern = pattern;
+	p->len = plen;
+
+	plen++;			/* Avoid increasing within the memset() inlined macro */
+
+	for (i = 0; i < ALPHA_SIZE; i++)
+		*pd++ = plen;
+
+	plen--;			/* Restore original pattern length */
+
+	for (pd = p->delta, c = (guchar *) pattern, i = 0; i < plen; c++, i++)
+		pd[*c] = plen - i;
+
+	return p;
+}
+
+/*
+ * entry_match
+ *
  * Apply pattern matching on text, matching at the *beginning* of words.
+ * Patterns are lazily compiled as needed, using pattern_compile_fast().
  */
 static gboolean entry_match(
 	gchar *text, gint tlen,
@@ -492,6 +689,10 @@ static gboolean entry_match(
 		gint amount = wovec[i].amount;
 		gint j;
 		guint32 offset = 0;
+
+		if (pw[i] == NULL)
+			pw[i] = pattern_compile_fast(wovec[i].word, wovec[i].len);
+
 		for (j = 0; j < amount; j++) {
 			char *pos =
 				pattern_qsearch(pw[i], text, tlen, offset, qs_begin);
@@ -510,9 +711,10 @@ static gboolean entry_match(
 /* do an actual search */
 gint st_search(
 	search_table_t *table,
-	guchar *search,
-	gboolean (*callback)(struct shared_file *),
-	gint max_res)
+	gchar *search,
+	gboolean (*callback)(shared_file_t *),
+	gint max_res,
+	query_hashvec_t *qhv)
 {
 	gint i, key, nres = 0;
 	guint len;
@@ -525,25 +727,14 @@ gint st_search(
 	gint vcnt;
 	gint scanned = 0;		/* measure search mask efficiency */
 	guint32 search_mask;
+	gint minlen;
 
-	len = map_string(table->fold_map, search);
-
-	/*
-	 * Prepare matching optimization, an idea from Mike Green.
-	 *
-	 * At library building time, we computed a mask hash, made from the
-	 * lowercased file name, using one bit per different letter, roughly
-	 * (see mask_hash() for the exact algorigthm).
-	 *
-	 * We're now going to compute the same mask on the query, and compare
-	 * it bitwise with the mask for each file.  If the file does not hold
-	 * at least all the chars present in the query, it's no use applying
-	 * the pattern matching algorithm, it won't match at all.
-	 *
-	 *		--RAM, 01/10/2001
-	 */
-
-	search_mask = mask_hash(search);
+#ifndef USE_ICU
+	len = match_map_string(table->fold_map, search);
+#else
+	len = strlen(search);
+	/* We don't need of the map because the strings are normalized */
+#endif
 
 	/*
 	 * Find smallest bin
@@ -551,7 +742,7 @@ gint st_search(
 
 	for (i = 0; i < len-1; i++) {
 		struct st_bin *bin;
-		if (isspace(search[i]) || isspace(search[i+1]))
+		if (isspace((guchar) search[i]) || isspace((guchar) search[i+1]))
 			continue;
 		key = st_key(table, search + i);
 		if ((bin = table->bins[key]) == NULL) {
@@ -577,22 +768,72 @@ gint st_search(
 	 *		--RAM, 06/10/2001
 	 */
 
-	if (best_bin == NULL)
-		return 0;
+	if (best_bin == NULL) {
+		/*
+		 * If we have a `qhv', we need to compute the word vector anway,
+		 * for query routing...
+		 */
 
-	g_assert(best_bin_size > 0);	/* Allocated bin, it must hold something */
+		if (qhv == NULL)
+			return 0;
+	}
 
 	/*
 	 * Prepare matching patterns
 	 */
 
 	wocnt = query_make_word_vec(search, &wovec);
-	if (wocnt == 0)
-		return 0;
-	pattern = (cpattern_t **) g_malloc(wocnt * sizeof(cpattern_t *));
 
-	for (i = 0; i < wocnt; i++)
-		pattern[i] = pattern_compile(wovec[i].word);
+	/*
+	 * Compute the query hashing information for query routing, if needed.
+	 */
+
+	if (qhv != NULL) {
+		for (i = 0; i < wocnt; i++) {
+			if (wovec[i].len >= QRP_MIN_WORD_LENGTH)
+				qhvec_add(qhv, wovec[i].word, QUERY_H_WORD);
+		}
+	}
+
+	if (wocnt == 0 || best_bin == NULL) {
+		if (wocnt > 0)
+			query_word_vec_free(wovec, wocnt);
+		return 0;
+	}
+
+	g_assert(best_bin_size > 0);	/* Allocated bin, it must hold something */
+
+
+	pattern = (cpattern_t **) g_malloc0(wocnt * sizeof(cpattern_t *));
+
+	/*
+	 * Prepare matching optimization, an idea from Mike Green.
+	 *
+	 * At library building time, we computed a mask hash, made from the
+	 * lowercased file name, using one bit per different letter, roughly
+	 * (see mask_hash() for the exact algorigthm).
+	 *
+	 * We're now going to compute the same mask on the query, and compare
+	 * it bitwise with the mask for each file.  If the file does not hold
+	 * at least all the chars present in the query, it's no use applying
+	 * the pattern matching algorithm, it won't match at all.
+	 *
+	 *		--RAM, 01/10/2001
+	 */
+
+	search_mask = mask_hash(search);
+
+	/*
+	 * Prepare second matching optimization: since all words in the query
+	 * must match the exact amount of time, we can compute the minimum length
+	 * the searched file must have.  We add one character after each word
+	 * but the last, to account for space between words.
+	 *		--RAM, 11/07/2002
+	 */
+
+	for (minlen = 0, i = 0; i < wocnt; i++)
+		minlen += wovec[i].len + 1;
+	minlen--;
 
 	/*
 	 * Search through the smallest bin
@@ -608,6 +849,9 @@ gint st_search(
 		if ((e->mask & search_mask) != search_mask)
 			continue;		/* Can't match */
 
+		if (sf->file_name_len < minlen)
+			continue;		/* Can't match */
+
 		scanned++;
 
 		if (entry_match(e->string, sf->file_name_len, pattern, wovec, wocnt)) {
@@ -627,27 +871,14 @@ gint st_search(
 			scanned, best_bin_size, nres);
 
 	for (i = 0; i < wocnt; i++)
-		pattern_free(pattern[i]);
-	g_free(pattern);
+		if (pattern[i])					/* Lazily compiled by entry_match() */
+			zfree(pat_zone, pattern[i]);
+
+	G_FREE_NULL(pattern);
 	query_word_vec_free(wovec, wocnt);
 
 	return nres;
 }
 
-/* minimize space consumption */
-void st_compact(search_table_t *table)
-{
-	gint i;
-
-	if (!table->all_entries.nvals)
-		return;			/* Nothing in table */
-
-	bin_compact(&table->all_entries);
-	for (i = 0; i < table->nbins; i++)
-		if (table->bins[i])
-			bin_compact(table->bins[i]);
-}
-
-
 /* vi: set ts=4: */
 
